import { Seq } from 'immutable';

export default class TreeUtils {

	constructor(idKey = 'id', typeKey = 'type', childNodesKey = 'childNodes') {
		this._idKey = idKey;
		this._childNodesKey = childNodesKey;
	}

	id(state, keyPath) {
		return state.getIn(keyPath.concat(this._idKey));
	}

	*__iterate(state, rootPath) {
		let stack = [rootPath];
		while (stack.length > 0) {
			let keyPath = stack.shift();
			yield keyPath;

			let item = state.getIn(keyPath),
				childNodes = item.get(this._childNodesKey);

			if (childNodes && childNodes.size > 0) {
				for (let i of item.get(this._childNodesKey).keys()) {
					stack.push(keyPath.concat(this._childNodesKey, i));
				}
			}
		}
	}

	findInTree(state, rootPath, comparator) {
		for (var keyPath of this.__iterate(state, rootPath)) {
			if (comparator(state.getIn(keyPath)) === true) {
				return keyPath;
			}
		}
	}

	filterInTree(state, rootPath, comparator) {
		let result = [];
		for (var keyPath of this.__iterate(state, rootPath)) {
			if (comparator(state.getIn(keyPath)) === true) {
				result.push(keyPath);
			}
		}
		return result;
	}

	getById(state, rootPath, id) {
		return this.findInTree(state, rootPath, (item) => {
			return item.get(this._idKey) === id;
		});
	}

	getByArbitrary(state, rootPath, idOrKeyPath) {
		return Seq.isSeq(idOrKeyPath) ?
			idOrKeyPath :
			this.getById(state, rootPath, idOrKeyPath);
	}

	nextSibling(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id),
			index = Number(keyPath.last()),
			nextSiblingPath = keyPath.skipLast(1).concat(index + 1);
		return state.hasIn(nextSiblingPath) ? nextSiblingPath : undefined;
	}

	previousSibling(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id),
			index = Number(keyPath.last());
		return index > 0 ? keyPath.skipLast(1).concat(index - 1) : undefined;
	}

	firstChild(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id).concat([this._childNodesKey, 0]);
		return state.hasIn(keyPath) ? keyPath : undefined;
	}

	lastChild(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id).concat([this._childNodesKey]),
			item = state.getIn(keyPath);
		if (item && item.size > 0) {
			return keyPath.concat([item.size - 1]);
		}

	}

	siblings(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id),
			index = Number(keyPath.last()),
			parentChildren = keyPath.skipLast(1),
			item = state.getIn(parentChildren);
		if (item) {
			let result = [];
			for (let i of item.keys()) {
				if (i !== index) {
					result.push(parentChildren.concat(i));
				}
			}
			return result;
		}
	}

	childNodes(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id).concat(this._childNodesKey),
			item = state.getIn(keyPath);
		if (item) {
			let l = item.size,
				result = [];
			for (let i = 0; i < l; i++) {
				result.push(keyPath.concat(i));
			}
			return result;
		}
	}


	childAt(state, rootPath, id, index) {
		let keyPath = this.getByArbitrary(state, rootPath, id).concat(this._childNodesKey, index);
		return state.hasIn(keyPath) ? keyPath : undefined;
	}


	descendants(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id);
		return this.filterInTree(state, keyPath, (item) => {
			return item.get(this._idKey) !== this.id(state, keyPath);
		});
	}

	childIndex(state, rootPath, id) {
		return Number(this.getByArbitrary(state, rootPath, id).last());
	}

	hasChildNodes(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id).concat(this._childNodesKey),
			item = state.getIn(keyPath);
		return typeof item !== 'undefined' && item.size > 0;
	}

	numChildNodes(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id).concat(this._childNodesKey),
			item = state.getIn(keyPath);
		return typeof item !== 'undefined' ? item.size : 0;
	}

	parent(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id);
		return keyPath && keyPath.size ? keyPath.slice(0, -2) : undefined;
	}

	ancestors(state, rootPath, id) {
		return this.getByArbitrary(state, rootPath, id).reduceRight((memo, value, index, keyPath) => {
			if ((index - rootPath.size) % 2 === 0 && index >= rootPath.size) {
				memo.push(
					new Seq(keyPath.takeLast(index).reverse().toArray())
				);
			}
			return memo;
		}, []);
	}

	right(state, rootPath, id) {
		let l = rootPath.size,
			keyPath = this.getByArbitrary(state, rootPath, id),
			firstChild = this.firstChild(state, rootPath, keyPath);

		if (firstChild) {
			return firstChild;
		}

		let nextSibling = this.nextSibling(state, rootPath, keyPath);
		if (nextSibling) {
			return nextSibling;
		}

		let parent = this.parent(state, rootPath, keyPath),
			nextSiblingOfParent;

		while(parent && parent.size >= l) {
			nextSiblingOfParent = this.nextSibling(state, rootPath, parent);
			if (nextSiblingOfParent) {
				return nextSiblingOfParent;
			}
			parent = this.parent(state, rootPath, parent);
		}

	}

	left(state, rootPath, id) {
		let keyPath = this.getByArbitrary(state, rootPath, id),
			lastChild = this.previousSibling(state, rootPath, keyPath);

		while (lastChild) {
			if (!this.hasChildNodes(state, rootPath, lastChild)) {
				return lastChild;
			}
			lastChild = this.lastChild(state, rootPath, lastChild);
		}

		let parent = this.parent(state, rootPath, keyPath);
		if(parent && parent.size >= rootPath.size) {
			return parent;
		}

	}

	firstDescendant(state, rootPath, id) {
		return this.firstChild(state, rootPath, id);
	}

	lastDescendant(state, rootPath, id) {
		let keyPath = this.lastChild(state, rootPath, id);
		while(keyPath && this.hasChildNodes(state, rootPath, keyPath)) {
			keyPath = this.lastChild(state, rootPath, keyPath);
		}
		return keyPath;
	}
}
