import Immutable from 'immutable';
import TreeAPI from '../src/TreeAPI';
import TreeCursor from '../src/TreeCursor';


describe('class `TreeCursor`', () => {
	let fixtures = {
			'data': {
				'name': 'Article',
				'type': 'article',
				'id': '1',
				'childNodes': [
					{
						'type': 'paragraph',
						'name': 'Paragraph',
						'id': '2'
					},
					{
						'type': 'list',
						'name': 'List',
						'id': '3',
						'childNodes': [
							{
								'type': 'listItem',
								'name': 'List item 1',
								'id': '4',
								'childNodes': [
									{
										'type': 'paragraph',
										'name': 'Nested paragraph',
										'id': '5'
									}
								]
							},
							{
								'type': 'listItem',
								'name': 'List item 2',
								'id': '6',
								'childNodes': [
									{
										'type': 'paragraph',
										'name': 'Nested paragraph 2',
										'id': '7'
									}
								]
							}
						]
					}
				]
			}
		},
		state = Immutable.fromJS(fixtures),
		api = new TreeAPI(),
		root;

		beforeEach(() => {
			root = api.treeFrom(state, ['data']);
		});

		describe('method `findInTree`', () => {
			it('returns the first node for which the comparator returned true.', () => {
				let result = root.getById('3').findInTree(item => item.get('name') === 'Nested paragraph 2');
				expect(result.id()).toEqual('7');
			});

			it('returns a TreeCursor.', () => {
				let result = root.getById('3').findInTree(item => item.get('name') === 'Nested paragraph 2');
				expect(result instanceof TreeCursor).toBeTruthy();
			});
		});

		describe('method `filterInTree`', () => {
			it('returns all nodes for which the comparator returned true.', () => {
				let result = root.getById('3').filterInTree(item => Number(item.get('id')) < 7).map(i => i.id());
				expect(result).toEqual([ '3', '4', '6', '5' ]);
			});

			it('returns a list of TreeCursors.', () => {
				let onlyTreeCursors = root.getById('3').filterInTree(item => Number(item.get('id')) < 7).every((cursor) => {
					return cursor instanceof TreeCursor;
				});
				expect(onlyTreeCursors).toBeTruthy();
			});
		});

		describe('method `id`', () => {
			it('returns the cursor\'s id.', () => {
				expect(
					root.getById('3').id()
				).toEqual('3');
			});
		});

		describe('method `parent`', () => {
			it('returns the parent element of the cursor.', () => {
				expect(root.getById('3').parent().id()).toEqual('1');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').parent() instanceof TreeCursor).toBeTruthy();
			});
		});

		describe('method `childIndex`', () => {
			it('returns the index of the cursor according to its parent list.', () => {
				expect(root.getById('3').childIndex()).toEqual(1);
			});
		});

		describe('method `previousSibling`', () => {
			it('returns the previous sibling node.', () => {
				expect(root.getById('3').previousSibling().id()).toEqual('2');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').previousSibling() instanceof TreeCursor).toBeTruthy();
			});
		});

		describe('method `nextSibling`', () => {
			it('returns the next sibling node.', () => {
				expect(root.getById('3').previousSibling().nextSibling().id()).toEqual('3');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').previousSibling().nextSibling() instanceof TreeCursor).toBeTruthy();
			});
		});

		describe('method `firstChild`', () => {
			it('returns the first child node.', () => {
				expect(root.getById('3').firstChild().id()).toEqual('4');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').firstChild() instanceof TreeCursor).toBeTruthy();
			});
		});

		describe('method `lastChild`', () => {
			it('returns the last child node.', () => {
				expect(root.getById('3').lastChild().id()).toEqual('6');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').lastChild() instanceof TreeCursor).toBeTruthy();
			});
		});

		describe('method `root`', () => {
			it('returns the root cursor.', () => {
				expect(root.getById('3').root().id()).toEqual('1');
			});
		});

		describe('method `hasChildNodes`', () => {
			it('returns true if a node has 1 or more child nodes.', () => {
				expect(root.getById('3').hasChildNodes()).toBeTruthy();
			});
		});

		describe('method `numChildNodes`', () => {
			it('returns the number of child nodes.', () => {
				expect(root.getById('3').numChildNodes('1')).toEqual(2);
			});
		});

		describe('method `siblings`', () => {
			it('returns list of all cursors with the same parent excluding the cursor itself.', () => {

				let ids = root.getById('3').siblings().map((cursor) => {
					return cursor.id();
				});
				expect(ids).toEqual(['2']);
			});

			it('returns a list of TreeCursor\'s.', () => {
				let onlyTreeCursors = root.getById('3').siblings().every((cursor) => {
					return cursor instanceof TreeCursor;
				});
				expect(onlyTreeCursors).toBeTruthy();
			});
		});

		describe('method `childNodes`', () => {
			it('returns list of all cursors which are first level descendants of the cursor.', () => {
				let ids = root.childNodes().map((cursor) => {
						return cursor.id();
					});
				expect(ids).toEqual(['2', '3']);
			});

			it('returns a list of TreeCursor\'s.', () => {
				let onlyTreeCursors = root.childNodes().every((cursor) => {
					return cursor instanceof TreeCursor;
				});
				expect(onlyTreeCursors).toBeTruthy();
			});
		});

		describe('method `ChildAt`', () => {
			it('returns the child cursor at `index`.', () => {
				expect(root.getById('3').childAt(1).id()).toEqual('6');
			});

			it('returns null if there is no child node at `index`.', () => {
				expect(root.getById('3').childAt(6)).toBeUndefined();
			});
		});

		describe('method `ancestors`', () => {
			it('returns a list of all ancestors of the node.', () => {
				let ids = root.getById('7').ancestors().map((cursor) => {
						return cursor.id();
					});
				expect(ids).toEqual(['6', '3', '1']);
			});

			it('returns a list of TreeCursor\'s.', () => {
				let onlyTreeCursors = root.getById('7').ancestors().every((cursor) => {
					return cursor instanceof TreeCursor;
				});
				expect(onlyTreeCursors).toBeTruthy();
			});
		});

		describe('method `descendants`', () => {
			it('returns a list of child node cursors.', () => {
				let ids = root.getById('3').descendants().map((cursor) => {
						return cursor.id();
					});
				expect(ids).toEqual([ '4', '6', '5', '7' ]);
			});

			it('returns a list of TreeCursor\'s.', () => {
				let onlyTreeCursors = root.getById('3').descendants().every((cursor) => {
					return cursor instanceof TreeCursor;
				});
				expect(onlyTreeCursors).toBeTruthy();
			});
		});

		describe('method `right`', () => {
			it('returns the next sibling of the cursor.', () => {
				expect(root.getById('3').right().id()).toEqual('4');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').right() instanceof TreeCursor).toBeTruthy();
			});

			it('can iterate over the whole tree.', () => {
				let id,
					node = root.getById('3'),
					result = [];
				while ((node = node.right()) !== undefined) {
					id = node.id();
					result.push(id);
				}
				expect(result).toEqual(['4', '5', '6', '7']);
			});
		});

		describe('method `left`', () => {
			it('returns the previous sibling of the node found at `id`', () => {
				expect(root.getById('3').left().id()).toEqual('2');
			});

			it('returns a TreeCursor.', () => {
				expect(root.getById('3').left() instanceof TreeCursor).toBeTruthy();
			});

			it('can iterate over the whole tree.', () => {
				let id,
					node = root.getById('3'),
					result = [];

				while ((node = node.left()) !== undefined) {
					id = node.id();
					result.push(id);
				}
				expect(result).toEqual(['2', '1']);
			});
		});

		describe('method `firstDescendant`', () => {
			it('returns the first child node in the tree', () => {
				expect(root.getById('3').firstDescendant().id()).toEqual('4');
			});
		});

		describe('method `lastDescendant`', () => {
			it('returns the last child node in the tree', () => {
				expect(root.getById('3').lastDescendant().id()).toEqual('7');
			});
		});
});
