import Immutable from 'immutable';
import TreeUtils from '../src/TreeUtils';


describe('class `TreeUtils`', () => {

	let fixtures = {
			'data': {
				'name': 'Article',
				'type': 'article',
				'id': '1',
				'childNodes': [
					{
						'type': 'paragraph',
						'name': 'Paragraph',
						'id': '2'
					},
					{
						'type': 'list',
						'name': 'List',
						'id': '3',
						'childNodes': [
							{
								'type': 'listItem',
								'name': 'List item 1',
								'id': '4',
								'childNodes': [
									{
										'type': 'paragraph',
										'name': 'Nested paragraph',
										'id': '5'
									}
								]
							},
							{
								'type': 'listItem',
								'name': 'List item 2',
								'id': '6',
								'childNodes': [
									{
										'type': 'paragraph',
										'name': 'Nested paragraph 2',
										'id': '7'
									}
								]
							}
						]
					}
				]
			}
		},
		state = Immutable.fromJS(fixtures),
		seq = Immutable.Seq,
		utils = new TreeUtils();

	describe('method `id`', () => {
		it('returns the id value of an absolute key path', () => {
			expect(utils.id(state, seq(['data', 'childNodes', 1]))).toEqual('3');
		});

		it('returns undefined if the key path has no id key.', () => {
			expect(utils.id(state, seq(['data', 'childNodes', 4]))).toBeUndefined();
		});
	});

	describe('method `findInTree`', () => {
		it('returns the first key path whose item passed the comparator function.', () => {
			let keyPath = utils.findInTree(state, seq(['data']), function(item) {
				return item.get('id') === '3';
			});
			expect(keyPath.toArray()).toEqual([ 'data', 'childNodes', 1]);
		});
	});

	describe('method `filterInTree`', () => {
		it('returns a list of all key paths whose items passed the comparator function.', () => {
			let keyPaths = utils.filterInTree(state, seq(['data']), function(item) {
				return item.get('type') === 'paragraph';
			}, []);
			expect(keyPaths.map(m => m.toArray())).toEqual([
				[ 'data', 'childNodes', 0 ],
				[ 'data', 'childNodes', 1, 'childNodes', 0, 'childNodes', 0 ],
				[ 'data', 'childNodes', 1, 'childNodes', 1, 'childNodes', 0 ]
			]);
		});
	});

	describe('method `getById`', () => {
		it('returns the key path of the first item whose id equals the id parameter.', () => {
			let keyPath = utils.getById(state, seq(['data']), '7');
			expect(keyPath.toArray()).toEqual([ 'data', 'childNodes', 1, 'childNodes', 1, 'childNodes', 0 ]);
		});
	});


	describe('method `getByArbitrary`', () => {
		it('returns the key path representing an id.', () => {
			expect(utils.getByArbitrary(state, seq(['data']), '4').toArray()).toEqual(
				['data', 'childNodes', 1, 'childNodes', 0]
			);
		});

		it('returns the key path parameter if it is an array / a key path already.', () => {
			expect(utils.getByArbitrary(state, seq(['data']), seq(['data', 'childNodes', 1, 'childNodes', 0])).toArray()).toEqual(
				['data', 'childNodes', 1, 'childNodes', 0]
			);
		});
	});

	describe('method `nextSibling`', () => {
		it('returns the next sibling node.', () => {
			let keyPath = utils.nextSibling(state, seq(['data']), '4');
			expect(keyPath.toArray()).toEqual(
				[ 'data', 'childNodes', 1, 'childNodes', 1]
			);
		});

		it('returns undefined if the node at `id` does not have a next sibling.', () => {
			let keyPath = utils.nextSibling(state, seq(['data']), '7');
			expect(keyPath).toBeUndefined();
		});
	});

	describe('method `previousSibling`', () => {
		it('returns the previous sibling node.', () => {
			let keyPath = utils.previousSibling(state, seq(['data']), '6');
			expect(keyPath.toArray()).toEqual(
				[ 'data', 'childNodes', 1, 'childNodes', 0]
			);
		});

		it('returns undefined if the node at `id` does not have a previous sibling.', () => {
			let keyPath = utils.previousSibling(state, seq(['data']), '4');
			expect(keyPath).toBeUndefined();
		});
	});

	describe('method `parent`', () => {
		it('returns the parent element\'s key path of the first item whose id equals the id parameter.', () => {
			let keyPath = utils.parent(state, seq(['data']), '7');
			expect(keyPath.toArray()).toEqual([ 'data', 'childNodes', 1, 'childNodes', 1]);
		});
	});

	describe('method `childIndex`', () => {
		it('returns the index of the first item whose id equals the id parameter according to its parent list.', () => {
			let index = utils.childIndex(state, seq(['data']), '3');
			expect(index).toEqual(1);
		});
	});

	describe('method `childAt`', () => {
		it('returns the composed keyPath that points at the child cursor.', () => {
			let keyPath = utils.childAt(state, seq(['data']), '3', 1);
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 1, 'childNodes', 1]);
		});

		it('returns undefined if the node at `id` has no child nodes or no child node at `index`.', () => {
			let keyPath = utils.childAt(state, seq(['data']), '2', 1);
			expect(keyPath).toBeUndefined();
		});
	});

	describe('method `firstChild`', () => {
		it('returns the first child node.', () => {
			let keyPath = utils.firstChild(state, seq(['data']), '4');
			expect(keyPath.toArray()).toEqual(
				['data', 'childNodes', 1, 'childNodes', 0, 'childNodes', 0]
			);
		});

		it('returns undefined if the node at `id` does not have a first child.', () => {
			let keyPath = utils.firstChild(state, seq(['data']), '2');
			expect(keyPath).toBeUndefined();
		});
	});

	describe('method `lastChild`', () => {
		it('returns the last child node.', () => {
			let keyPath = utils.lastChild(state, seq(['data']), '1');
			expect(keyPath.toArray()).toEqual(
				['data', 'childNodes', 1]
			);
		});

		it('returns undefined if the node at `id` does not have a last child.', () => {
			let keyPath = utils.lastChild(state, seq(['data']), '7');
			expect(keyPath).toBeUndefined();
		});
	});

	describe('method `hasChildNodes`', () => {
		it('returns true if a node has 1 or more child nodes.', () => {
			expect(utils.hasChildNodes(state, seq(['data']), '4')).toBeTruthy();
		});

		it('returns false if a node has no child nodes.', () => {
			expect(utils.hasChildNodes(state, seq(['data']), '5')).toBeFalsy();
		});
	});

	describe('method `numChildNodes`', () => {
		it('returns the number of child nodes.', () => {
			expect(utils.numChildNodes(state, seq(['data']), '1')).toEqual(2);
		});
	});

	describe('method `siblings`', () => {
		it('returns list of all key paths with the same parent excluding the node itself.', () => {
			let keyPaths = utils.siblings(state, seq(['data']), '6');
			expect(keyPaths.map(m => m.toArray())).toEqual([
				[ 'data', 'childNodes', 1, 'childNodes', 0 ]
			]);
		});
	});

	describe('method `childNodes`', () => {
		it('returns list of all key paths which are first level descendants of the given node.', () => {
			let keyPaths = utils.childNodes(state, seq(['data']), '1');
			expect(keyPaths.map(m => m.toArray())).toEqual([
				[ 'data', 'childNodes', 0 ],
				[ 'data', 'childNodes', 1 ]
			]);
		});
	});

	describe('method `ancestors`', () => {
		it('returns a list of all key paths that point at the ancestor nodes of the item at `id`.', () => {
			let keyPaths = utils.ancestors(state, seq(['data']), '7');
			expect(keyPaths.map(m => m.toArray())).toEqual([
				[ 'data', 'childNodes', 1, 'childNodes', 1],
				[ 'data', 'childNodes', 1],
				[ 'data']
			]);
		});
	});

	describe('method `descendants`', () => {
		it('returns a list of all key paths that represent child nodes of the item at `id`', () => {
			let keyPaths = utils.descendants(state, seq(['data']), '3');
			expect(keyPaths.map(m => m.toArray())).toEqual([
				[ 'data', 'childNodes', 1, 'childNodes', 0 ],
				[ 'data', 'childNodes', 1, 'childNodes', 1 ],
				[ 'data', 'childNodes', 1, 'childNodes', 0, 'childNodes', 0 ],
				[ 'data', 'childNodes', 1, 'childNodes', 1, 'childNodes', 0 ]
			]);
		});
	});

	describe('method `right`', () => {
		it('returns the next sibling of the node found at `id`', () => {
			let keyPath = utils.right(state, seq(['data']), '2');
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 1]);
		});

		it('returns the first child node if present.', () => {
			let keyPath = utils.right(state, seq(['data']), '3');
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 1, 'childNodes', 0]);
		});

		it('returns the next sibling of the first ancestor that has a next sibling', () => {
			let keyPath = utils.right(state, seq(['data']), '5');
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 1, 'childNodes', 1]);
		});

		it('returns undefined if it is the last node in the tree.', () => {
			let keyPath = utils.right(state, seq(['data']), '7');
			expect(keyPath).toBeUndefined();
		});

		it('can iterate over the whole tree.', () => {
			let id = '1',
				node,
				result = [];
			while ((node = utils.right(state, seq(['data']), id))) {
				id = utils.id(state, node);
				result.push(id);
			}
			expect(result).toEqual(['2', '3', '4', '5', '6', '7']);
		});
	});

	describe('method `left`', () => {
		it('returns the previous sibling of the node found at `id`', () => {
			let keyPath = utils.left(state, seq(['data']), '3');
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 0]);
		});

		it('returns the latest child of the previous sibling if exists.', () => {
			let keyPath = utils.left(state, seq(['data']), '6');
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 1, 'childNodes', 0, 'childNodes', 0]);
		});

		it('returns the parent node if the node at `id` has no previous sibling.', () => {
			let keyPath = utils.left(state, seq(['data']), '7');
			expect(keyPath.toArray()).toEqual(['data', 'childNodes', 1, 'childNodes', 1]);
		});

		it('returns undefined if it is the last node in the tree.', () => {
			let keyPath = utils.left(state, seq(['data']), '1');
			expect(keyPath).toBeUndefined();
		});

		it('can iterate over the whole tree.', () => {
			let id = '7',
				node,
				result = [];

			while ((node = utils.left(state, seq(['data']), id))) {
				id = utils.id(state, node);
				result.push(id);
			}
			expect(result).toEqual(['6', '5', '4', '3', '2', '1']);
		});
	});


	describe('method `firstDescendant`', () => {
		it('returns the first child node in the tree', () => {
			let keyPath = utils.firstDescendant(state, seq(['data']), '1');
			expect(keyPath.toArray()).toEqual([ 'data', 'childNodes', 0]);
		});
	});

	describe('method lastDescendant`', () => {
		it('returns the last child node in the tree', () => {
			let keyPath = utils.lastDescendant(state, seq(['data']), '1');
			expect(keyPath.toArray()).toEqual([ 'data', 'childNodes', 1, 'childNodes', 1, 'childNodes', 0]);
		});
	});
});
